//Wrote by Kevin Maingault, last update 08/06/2024. Kevin.maingault@gmail.com
// Standardized Tile 32*32 version

//we declare var that will be used for calculation, we declare var here instead inside the function to dont recreate them in memory each time the function is called
var referenceDate = new Date("2024-01-01T00:00:00.000+00:00");  // date as Year, month, day, hours, minutes, secondes, reference date of the following value, as UTC specified to be sure there is no error

//Phase value for the reference date for the used constituant. idealy these datas need to be updated to take in consideration some variaiton of moon and sun position
//for this project updating it once a year is enough to get something very accurate. but same withotu updating it ti will stay accurate enough on half hour prediction for 20 years and more. 
var K1P92 = 0.128103531;
var M2P92 = 4.319695129;
var N2P92 = 1.245078394;
var O1P92 = 4.211522846;
var S2P92 = 0;

//Pulsation Radian/minutes (fixed constant, = period sine curve)
var K1T = 0.0043752702;
var M2T = 0.008431134;
var N2T = 0.008272782;
var O1T = 0.0040558644;
var S2T = 0.008726646;


function getTileX(lon, zoom) { //function to convert longitude and zoom to X tile coorinate  //Read OSM wiki for furthe rinformation
    let xtile = parseInt(Math.floor((lon + 180) / 360 * (1 << zoom)));
    return xtile ;
}

function getTileY(lat, zoom) { //function to convert latitude and zoom to y tile coorinate
    let ytile = parseInt(Math.floor((1 - Math.log(Math.tan((lat * Math.PI / 180)) + 1 / Math.cos((lat * Math.PI / 180))) / Math.PI) / 2 * (1 << zoom)));
    return ytile;
}

function xToLon(x, zoom) { //function to convert X tile to Longitude
    let n = 2**zoom;
    let lon_deg = x / n * 360 - 180;
    return lon_deg ;
}

function yToLat(y, zoom) { //function to convert Y tile to lattitude
    let n = 2**zoom;
    let lat_rad = Math.atan(Math.sinh(Math.PI * (1 - 2 * y / n)));
    let lat_deg = lat_rad * 180 / Math.PI;
    return lat_deg;
}

function dateAddHours(a, b) {
    var d = new Date(b || new Date());
    d.setHours(d.getHours() + a);
    return d;
  }


  function participation(harmonic, xPos, yPos, i) {     //Function to calcul participation from harmonics value. Rewrotte from my original work under c# app

    let xPix = (xPos * 32) + i - (Math.floor(i / 32) * 32); //We redefine the X pixel position in the global Grid
    let yPix = (yPos * 32) + Math.floor(i / 32); //We redefine the Y pixel position in the global Grid

    harmonic = harmonic.split(',');    // return the string after splitting it by line
    let today = currentTime; //return nuber of millisecond UTC since 1960/epoch time, use date delcared in the script.js
    //today = dateAddHours(2, today);
    let minutes = Math.floor((today.valueOf() - referenceDate.valueOf()) / 1000 / 60) ;// use calueOF() to be sure to get the millisecond UTC since epoch date

    if (harmonic[0] === "L") {  // we do this to return 0 on no value data, identified by "L" in the tile file

        return [0, 0, xPix, yPix]
    }

    else {
        //calcul U value
        let K1u = harmonic[0] * Math.cos(K1P92 + (minutes * K1T) - (harmonic[1] / 180 * Math.PI));    //WARNINNNNGGGGGG Check well the position of parameter, it is ordered by constituent instead U and V
        let M2u = harmonic[4] * Math.cos(M2P92 + (minutes * M2T) - (harmonic[5] / 180 * Math.PI));     //I lost days of work deging it... stupid choice but too late to change it
        let N2u = harmonic[8] * Math.cos(N2P92 + (minutes * N2T) - (harmonic[9] / 180 * Math.PI));
        let O1u = harmonic[12] * Math.cos(O1P92 + (minutes * O1T) - (harmonic[13] / 180 * Math.PI));
        let S2u = harmonic[16] * Math.cos(S2P92 + (minutes * S2T) - (harmonic[17] / 180 * Math.PI));

        let u = (K1u + M2u + N2u + O1u + S2u); 


        //calcul v value
        let K1v = harmonic[2] * Math.cos(K1P92 + (minutes * K1T) - (harmonic[3] / 180 * Math.PI));
        let M2v = harmonic[6] * Math.cos(M2P92 + (minutes * M2T) - (harmonic[7] / 180 * Math.PI));
        let N2v = harmonic[10] * Math.cos(N2P92 + (minutes * N2T) - (harmonic[11] / 180 * Math.PI));
        let O1v = harmonic[14] * Math.cos(O1P92 + (minutes * O1T) - (harmonic[15] / 180 * Math.PI));
        let S2v = harmonic[18] * Math.cos(S2P92 + (minutes * S2T) - (harmonic[19] / 180 * Math.PI));

        let v = (K1v + M2v + N2v + O1v + S2v); 

        return [u, v, xPix, yPix]

    }
}

const velocityTile = async(url) => {  //function asynchrone that will wait to load all the txt before continuing process / returning the content/// catch url as param, string
    
    let response = await fetch (url[0]);  // initialize a variable response, let mean this variable is limited to this function. await fetch, mean we await to fully read the file specified by url to considere the var full
                                          // Fetch only catch networ error, thats mean if file is not avaible, it will not crash the app only the process. by the way if tile doesnt exist it just dont load it on the map
                                          // And like our main grid is filled with 0 by default, other datas are correctly loaded.
    
    const txt = await response.text().then(( str ) => {
        return str.split('\r\n');    // return the string after splitting it by line as an array
    });
    
    let parti = new Array();  // creating a new array to add harmonics datas
    let i = 1;
    while (i < txt.length)
    {
        if (txt[0] == "USA") { // here we check the first line of the tile, this could be used later to use different way to calcul participation
            parti[i-1] = participation(txt[i], url[1], url[2], (i-1));  //for each line we calcul and store the participation result
            i++;
        }

        else { // here we check the first line of the tile, this could be used later to use different way to calcul participation
          parti[i-1] = participation(txt[i], url[1], url[2], (i-1));  //for each line we calcul and store the participation result
            i++;
        }
    }
    
    return parti //we return all the results as an Array
}



function jsonMerge(west, north, south, east, zoom) {  // our main function that will create the Json file to feed Leaflet velocity (derived itsel from wind-essri derived itself from earth.nullschool)
    // This global If is used to avoid a double layer error. Like JS work with Synchronous function. If user zoom/pan too speed, a layer can be created before to be removed, resulting in a double veloity layer.
    //The good solution will be probably to rewrite the velocity layer loading as Asynchronous function but i'm tired to deal with this.
    zoom = zoom+1;
    let xtotal = getTileX(east, zoom) - getTileX(west, zoom) +1; //number of tile on X axis
    let ytotal = getTileY(south, zoom) - getTileY(north, zoom) +1; //number of tile en Y axis
    let xsize = xToLon(getTileX(east, zoom) + 1, zoom) - xToLon(getTileX(west, zoom), zoom); //size x total of the tile in degree coordinate
    let ysize = yToLat(getTileY(north, zoom), zoom) - yToLat(getTileY(south, zoom) + 1, zoom) ; //size y total of the tile in degree coordinate
    let rx = xsize / xtotal / 32;  //x cell resolution degree cordinate
    let ry = ysize / ytotal / 32;  //y cell resolution degree cordinate 
    let xarray = xtotal * 32; //size x of the array
    let yarray = ytotal * 32; // size y of the array 


    let jNorth = yToLat(getTileY(north, zoom), zoom); //Top left point coordinate
    let jWest = xToLon(getTileX(west, zoom), zoom);  //Top left point coordinate 

    //THE BIGGEST S**T with JS when you come from C#. Multi dimensional array are array in array. By the way you have two possibility, just initialising an array in an array and managing the size dynamically
    // but for large datas this will result in longer processing time than using a initialised array. The problem is like it is array in array you cant initiate a 2D or 3D aray in one time 
    // you need to initiate a first array at specified lenght, the 1D and then initiate each array insied field of this array. And if you want to fill it wil vallue or a third array (3D) you need to do it manually
    //if you use the .fill property, all value inside are linked to a single memory adress and if you try to update one value, all the vallue will be updated. It realy look like a bug, i dont see interrest for this feature else
    //like we want that empty value are 0 we have to fill it manualy and the size of this predefined array will be always a little larger than what we realy use. but based on benchmark, i think the win of time for large
    //for large array processing is better. Look here :  https://medium.datadriveninvestor.com/how-fast-is-javascript-to-c-or-c-29d104f4f255

    let jarray = new Array(xarray);  //we initiate first array / first dimension

    let i = 0;
    while (i < xarray) {jarray[i] = new Array(yarray); i++}  // we initiate array in each array / 2nd dimension
    i = 0;
    let j = 0;
    while (i < xarray) {
        while (j < yarray) {
            jarray[i][j] = [0 , 0];   // we want to fill with 0, in the same time we fill each field with a new array containing 0, 0 / the 3rd dimension to get u and v in the same array
            j++;
        }
    j = 0;
    i++;
    } 

    let xcounter = 0; //initiate few counters for following processing
    let ycounter = 0;
    let icounter = 0;
    let eachcounter = 0;

    //javascript in synchronous function based. realy not good to manage calculation after calculation. have to use Asynchrone function that look weird especialy because i'm not a good JS developper
    url = new Array();

    while ((getTileY(north, zoom) + ycounter) <= getTileY(south, zoom))  //we create loop in loop to process each tile on X and Y axis
    {
        while ((getTileX(west, zoom) + xcounter) <= getTileX(east, zoom)) //we create loop in loop to process each tile on X and Y axis
        {
            // here we initiate all the url of requested tile to execute all process after and in the same way we check these file are avaible. this check will help to manage no data avaible and avoir crash from
            url[icounter] = [tilePath + zoom + "/" + (getTileX(west, zoom) + xcounter)  + "/" + (getTileY(north, zoom) + ycounter) + ".csv", xcounter, ycounter]  //We send the tile position as param to reuse in participation
            xcounter++;
            icounter++;
        }
        xcounter = 0;
        ycounter++;
    }



    url.forEach(async (urli) => {  //we process each url to get each tile                                                // OK it seem all url inside ar good

      let result = await velocityTile(urli);   //load the tile                                                           //it seem it work, at less X tile content is different

       result.forEach(async (resulti) => { // we process each point from each file to order it in the main array
            if (resulti[0] != '' && resulti[0] != undefined && !isNaN(resulti[0])) //one more ugly thing to jump the last blank line from CSV file due to c# Writeline feature
            {
                let ax = resulti[2]; // get position X of the data based on first top corner point loaded before    //Will change once tile size standardized        
                let ay = resulti[3]; // get position Y of the data based on first top corner point loaded before    //Will change once tile size standardized
                jarray[ax][ay][0] = resulti[0];
                jarray[ax][ay][1] = resulti[1];
                
            }
            
        });
      
      
        eachcounter++;

        if (eachcounter === url.length) { //well, when we arrive here that mean all the ugly async functions was correctly executed before and we have now a jarray correctly filled with U V datas created on the fly from tile. Now it is time to convert it to json and send it to velocity layer.
        //we convert the datas from array to ascii json string. Thats stupid because velocity layer will reconvert it to array again. but for begining this is the fastest way to get something working
        var header1 = "[{\"header\": {\"parameterUnit\": \"m.s - 1\", \"parameterNumber\": 2, \"dx\":  " + rx + ", \"dy\": " + ry + ", \"parameterNumberName\": \"Eastward current\", \"la1\": " + jNorth + ", \"parameterCategory\": 2, \"nx\": " + xarray + ", \"ny\": " + yarray + ", \"refTime\": \"2017 - 02 - 01 23:00:00\", \"lo1\": " + jWest + "}, \"data\": [";
        var header2 = "]}, { \"header\": { \"parameterUnit\": \"m.s-1\", \"parameterNumber\": 3, \"dx\": " + rx + ", \"dy\": " + ry + ", \"parameterNumberName\": \"Northward current\", \"la1\": " + jNorth + ", \"parameterCategory\": 2, \"nx\": "+ xarray +", \"ny\": " + yarray + ", \"refTime\": \"2017-02-01 23:00:00\", \"lo1\": " + jWest + "}, \"data\": [";
        var footer = "]}]";
        var arrayU ="";
        var arrayV ="";

        let countx = 0; //like every time we initialise some counter for while
        let county = 0;
        while (county < yarray ) { //an important thing here, do not use string = string + data; it is not optimized, instead we use string += data;
            while (countx < xarray ) {
                if (countx === 0 && county === 0 ) { //no coma for the first value
                    arrayU +=  jarray[countx][county][0];
                    arrayV +=  jarray[countx][county][1];
                }
                else {
                    arrayU += ", " + jarray[countx][county][0];
                    arrayV += ", " + jarray[countx][county][1];
                }

            countx++;
        }
            countx = 0;
            county++;
        }

        let data = header1 + arrayU + header2 + arrayV + footer; //merge the data to create the json string
        let jdata = JSON.parse(data); //parse the data to get a json object
        loadVelocityLayer(jdata); //reload the velocity layer with the new datas
      }
    });
}